@@ -123,6 +123,7 @@ gboolean              nautilus_burn_drive_lock                     (NautilusBurn
 								    char                   **reason_for_failure);
 gboolean              nautilus_burn_drive_unlock                   (NautilusBurnDrive       *drive);
 gboolean              nautilus_burn_drive_unmount                  (NautilusBurnDrive       *drive);
+gboolean              nautilus_burn_drive_is_mounted               (NautilusBurnDrive       *drive);
 gboolean              nautilus_burn_drive_eject                    (NautilusBurnDrive       *drive);
 
 gboolean              nautilus_burn_drive_door_is_open             (NautilusBurnDrive       *drive);
@@ -135,9 +136,9 @@ gboolean              nautilus_burn_drive_can_rewrite              (NautilusBurn
 
 int                   nautilus_burn_drive_get_max_speed_write      (NautilusBurnDrive       *drive);
 int                   nautilus_burn_drive_get_max_speed_read       (NautilusBurnDrive       *drive);
+const int *           nautilus_burn_drive_get_write_speeds         (NautilusBurnDrive       *drive);
 
-
-/* Media type handling */
+/* Media handling */
 
 NautilusBurnMediaType nautilus_burn_drive_get_media_type           (NautilusBurnDrive       *drive);
 
@@ -147,11 +148,13 @@ NautilusBurnMediaType nautilus_burn_drive_get_media_type_full      (NautilusBurn
 								    gboolean                *has_data,
 								    gboolean                *has_audio);
 gint64                nautilus_burn_drive_get_media_size           (NautilusBurnDrive       *drive);
-
+char *                nautilus_burn_drive_get_media_label          (NautilusBurnDrive       *drive);
 
 const char *          nautilus_burn_drive_media_type_get_string    (NautilusBurnMediaType    type);
 char *                nautilus_burn_drive_get_supported_media_string (NautilusBurnDrive       *drive,
 								      gboolean                 writable_only);
+gboolean              nautilus_burn_drive_media_type_is_writable   (NautilusBurnMediaType    type,
+								    gboolean                 is_blank);
 
 /* The following are deprecated */
 NautilusBurnMediaType nautilus_burn_drive_get_media_type_from_path (const char              *device_path);
