@@ -3,7 +3,7 @@
  * nautilus-cd-burner.c: easy to use cd burner software
  * 
  * Copyright (C) 2002-2004 Red Hat, Inc.
- * Copyright (C) 2005 William Jon McCann <mccann@jhu.edu>
+ * Copyright (C) 2005-2006 William Jon McCann <mccann@jhu.edu>
  *   
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -25,6 +25,8 @@
  *          William Jon McCann <mccann@jhu.edu>
  */
 
+#include "config.h"
+
 #include <time.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -34,10 +36,9 @@
 #include <glib.h>
 #include <glib/gstdio.h>
 #include <gtk/gtk.h>
-#include <gtk/gtkfilesel.h>
-#include <gtk/gtkseparatormenuitem.h>
 #include <glade/glade.h>
 #include <libgnomevfs/gnome-vfs.h>
+
 #include <libgnome/gnome-help.h>
 #include <libgnomeui/gnome-ui-init.h>
 #include <gconf/gconf-client.h>
@@ -58,6 +59,19 @@ enum {
 	CANCEL_CD_RECORD,
 };
 
+enum {
+	NCB_SOURCE_BURN = 0,
+	NCB_SOURCE_IMAGE = 1,
+	NCB_SOURCE_DRIVE = 2,
+};
+
+enum {
+	NCB_OUTPUT_DEFAULT = 0,
+	NCB_OUTPUT_FILE = 1,
+	NCB_OUTPUT_CD = 2,
+	NCB_OUTPUT_DVD = 3,
+};
+
 typedef struct {
 	char              *source_device;
 	char              *source_iso;
@@ -86,13 +100,65 @@ static NautilusBurnerWriteData *write_data = NULL;
 
 static int write_disc (NautilusBurnerWriteData *data);
 
+/* Profiling stuff adapted from gtkfilechooserdefault */
+/* To use run:
+ *  strace -ttt -f -o logfile.txt nautilus-cd-burner
+ */
+
+#undef PROFILE_NCB
+#ifdef PROFILE_NCB
+
+#define PROFILE_INDENT 4
+static int profile_indent;
+
+static void
+profile_add_indent (int indent)
+{
+        profile_indent += indent;
+        if (profile_indent < 0)
+                g_error ("You screwed up your indentation");
+}
+
+static void
+_ncb_profile_log (const char *func,
+		  int         indent,
+		  const char *msg1,
+		  const char *msg2)
+{
+        char *str;
+
+        if (indent < 0)
+                profile_add_indent (indent);
+
+        if (profile_indent == 0)
+                str = g_strdup_printf ("MARK: %s: %s %s %s", G_STRLOC, func, msg1 ? msg1 : "", msg2 ? msg2 : "");
+        else
+                str = g_strdup_printf ("MARK: %s: %*c %s %s %s", G_STRLOC, profile_indent - 1, ' ', func, msg1 ? msg1 : "", msg2 ? msg2 : "");
+
+        access (str, F_OK);
+
+        g_free (str);
+
+        if (indent > 0)
+                profile_add_indent (indent);
+}
+
+#define profile_start(x, y) _ncb_profile_log (G_STRFUNC, PROFILE_INDENT, x, y)
+#define profile_end(x, y)   _ncb_profile_log (G_STRFUNC, -PROFILE_INDENT, x, y)
+#define profile_msg(x, y)   _ncb_profile_log (NULL, 0, x, y)
+#else
+#define profile_start(x, y)
+#define profile_end(x, y)
+#define profile_msg(x, y)
+#endif
+
 static void
 nautilus_burn_progress_write_data_free (NautilusBurnerWriteData *data)
 {
 	if (! data)
 		return;
 
-	nautilus_burn_drive_free (data->drive);
+	nautilus_burn_drive_unref (data->drive);
 
 	g_list_foreach (data->tracks, (GFunc)nautilus_burn_recorder_track_free, NULL);
 	g_list_free (data->tracks);
@@ -455,74 +521,18 @@ progress_changed_cb (NautilusBurnRecorder *recorder,
 	nautilus_burn_progress_set_time (secs);
 }
 
-static char *
-add_desc (char       *string,
-	  const char *addition)
-{
-	char *new;
-
-	if (strcmp (string, "") == 0) {
-		new = g_strdup_printf ("%s", addition);
-	} else {
-		new = g_strdup_printf ("%s/%s", string, addition);
-	}
-	g_free (string);
-	return new;
-}
-
-static char *
-drive_type_get_supported_media (int type)
-{
-	char *string;
-
-	string = g_strdup ("");
-	if (type & NAUTILUS_BURN_DRIVE_TYPE_CD_RECORDER) {
-		string = add_desc (string, "CD-R");
-	}
-	if (type & NAUTILUS_BURN_DRIVE_TYPE_CDRW_RECORDER) {
-		string = add_desc (string, "CD-RW");
-	}
-	if (type & NAUTILUS_BURN_DRIVE_TYPE_DVD_RAM_RECORDER) {
-		string = add_desc (string, "DVD-RAM");
-	}
-	if (type & NAUTILUS_BURN_DRIVE_TYPE_DVD_RW_RECORDER) {
-		string = add_desc (string, "DVD-RW");
-	}
-	if (type & NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_R_RECORDER) {
-		string = add_desc (string, "DVD+R");
-	}
-	if (type & NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_R_DL_RECORDER) {
-		string = add_desc (string, "DVD+R DL");
-	}
-	if (type & NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_RW_RECORDER) {
-		string = add_desc (string, "DVD+RW");
-	}
-
-	return string;
-}
-
-static gboolean
-drive_can_rewrite (const NautilusBurnDrive *drive)
-{
-	int type;
-
-	type = drive->type;
-
-	return (type & NAUTILUS_BURN_DRIVE_TYPE_CDRW_RECORDER
-		|| type & NAUTILUS_BURN_DRIVE_TYPE_DVD_RW_RECORDER
-		|| type & NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_RW_RECORDER);
-}
-
 static int  
-ask_for_media (const NautilusBurnDrive *drive,
-	       gboolean                 is_reload,
-	       gboolean                 busy_cd,
-	       gboolean                 ask_overburn,
-	       gint64                   required_size)
+ask_for_media (NautilusBurnDrive *drive,
+	       gboolean           is_reload,
+	       gboolean           busy_cd,
+	       gboolean           ask_overburn,
+	       gint64             required_size)
 {
 	GtkWidget  *dialog;
 	GtkWidget  *parent;
+	GtkWidget  *button;
 	gboolean    can_rewrite;
+	gboolean    show_eject;
 	char       *msg;
 	char       *type_string;
 	const char *title;
@@ -531,32 +541,35 @@ ask_for_media (const NautilusBurnDrive *drive,
 
 	mibs = (required_size + 1024 * 1024 - 1) / (1024 * 1024);
 
-	type_string = drive_type_get_supported_media (drive->type);
+	type_string = nautilus_burn_drive_get_supported_media_string (drive, TRUE);
 
-	can_rewrite = drive_can_rewrite (drive);
+	can_rewrite = nautilus_burn_drive_can_rewrite (drive);
+	show_eject = FALSE;
 
 	if (busy_cd) {
 		msg = g_strdup (_("Please make sure another application is not using the drive."));
 		title = N_("Drive is busy");
 	} else if (!is_reload && can_rewrite) {
-		msg = g_strdup_printf (_("Please put a %s, with at least %d MiB free, into the drive."),
-				       type_string,
-				       mibs);
+		msg = g_strdup_printf (_("Please put a disc, with at least %d MiB free, into the drive.  The following disc types are supported:\n%s"),
+				       mibs,
+				       type_string);
 		title = N_("Insert a rewritable or blank disc");
 	} else if (!is_reload && !can_rewrite) {
-		msg = g_strdup_printf (_("Please put a %s, with at least %d MiB free, into the drive."),
-				       type_string,
-				       mibs);
+		msg = g_strdup_printf (_("Please put a disc, with at least %d MiB free, into the drive.  The following disc types are supported:\n%s"),
+				       mibs,
+				       type_string);
 		title = N_("Insert a blank disc");
 	} else if (can_rewrite) {
-		msg = g_strdup_printf (_("Please replace the disc in the drive with a %s, with at least %d MiB free."),
-				       type_string,
-				       mibs);
+		show_eject = TRUE;
+		msg = g_strdup_printf (_("Please replace the disc in the drive with a supported disc with at least %d MiB free.  The following disc types are supported:\n%s"),
+				       mibs,
+				       type_string);
 		title = N_("Reload a rewritable or blank disc");
 	} else {
-		msg = g_strdup_printf (_("Please replace the disc in the drive with a %s, with at least %d MiB free."),
-				       type_string,
-				       mibs);
+		show_eject = TRUE;
+		msg = g_strdup_printf (_("Please replace the disc in the drive with a supported disc with at least %d MiB free.  The following disc types are supported:\n%s"),
+				       mibs,
+				       type_string);
 		title = N_("Reload a blank disc");
 	}
 
@@ -569,13 +582,26 @@ ask_for_media (const NautilusBurnDrive *drive,
 	dialog = ncb_hig_dialog (GTK_MESSAGE_INFO,
 				 _(title), _(msg), GTK_WINDOW (parent));
 
+	if (show_eject) {
+		button = gtk_dialog_add_button (GTK_DIALOG (dialog),
+						_("_Eject"),
+						GTK_RESPONSE_NO);
+	}
+
 	gtk_dialog_add_button (GTK_DIALOG (dialog), GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL);
 	gtk_dialog_add_button (GTK_DIALOG (dialog), GTK_STOCK_OK, GTK_RESPONSE_OK);
 
 	gtk_dialog_set_default_response (GTK_DIALOG (dialog),
 					 GTK_RESPONSE_OK);
 
+ retry:
 	res = gtk_dialog_run (GTK_DIALOG (dialog));
+
+	if (res == GTK_RESPONSE_NO) {
+		nautilus_burn_drive_eject (drive);
+		goto retry;
+	}
+
 	gtk_widget_destroy (dialog);
 
 	g_free (msg);
@@ -585,8 +611,8 @@ ask_for_media (const NautilusBurnDrive *drive,
 
 typedef struct
 {
-	gint64                   size;
-	const NautilusBurnDrive *drive;
+	gint64             size;
+	NautilusBurnDrive *drive;
 } InsertMediaCallbackData;
 
 static gboolean
@@ -737,6 +763,34 @@ file_info_is_allowed (GnomeVFSFileInfo *info)
 	return TRUE;
 }
 
+static char *
+get_backing_file (const char *virtual_path)
+{
+	GnomeVFSHandle *handle;
+	GnomeVFSResult  res;
+	char           *escaped_path, *uri;
+	char           *mapping;
+
+	escaped_path = gnome_vfs_escape_path_string (virtual_path);
+	uri = g_strconcat ("burn:///", escaped_path, NULL);
+	g_free (escaped_path);
+	/* warning, this can hang on fifos etc */
+	res = gnome_vfs_open (&handle,
+			      uri,
+			      GNOME_VFS_OPEN_READ);
+	g_free (uri);
+	if (res == GNOME_VFS_OK) {
+		res =  gnome_vfs_file_control (handle,
+					       "mapping:get_mapping",
+					       &mapping);
+		gnome_vfs_close	(handle);
+		if (res == GNOME_VFS_OK) {
+			return mapping;
+		}
+	}
+	return NULL;
+}
+
 static gboolean
 estimate_size_callback (const char       *rel_path,
 			GnomeVFSFileInfo *info,
@@ -747,7 +801,7 @@ estimate_size_callback (const char       *rel_path,
 	EstimateSizeCallbackData *cb_data = data;
 
 	if (file_info_is_image (info)) {
-		cb_data->image_paths = g_list_append (cb_data->image_paths, g_strdup (rel_path));
+		cb_data->image_paths = g_list_append (cb_data->image_paths, get_backing_file (rel_path));
 	} else {
 		cb_data->only_images = FALSE;
 	}
@@ -774,6 +828,8 @@ estimate_size (const char *uri,
 	GnomeVFSFileSize          size;
 	EstimateSizeCallbackData *data;
 
+	profile_start ("start", NULL);
+
 	data = g_new0 (EstimateSizeCallbackData, 1);
 	data->only_images = TRUE;
 
@@ -801,8 +857,9 @@ estimate_size (const char *uri,
 		if (unreadable_paths) {
 			GList *l;
 			*unreadable_paths = NULL;
-			for (l = data->unreadable_paths; l; l = l->next)
+			for (l = data->unreadable_paths; l; l = l->next) {
 				*unreadable_paths = g_list_append (*unreadable_paths, g_strdup ((char *)l->data));
+			}
 		}
 
 		g_list_foreach (data->unreadable_paths, (GFunc)g_free, NULL);
@@ -813,8 +870,9 @@ estimate_size (const char *uri,
 		if (image_paths) {
 			GList *l;
 			*image_paths = NULL;
-			for (l = data->image_paths; l; l = l->next)
+			for (l = data->image_paths; l; l = l->next) {
 				*image_paths = g_list_append (*image_paths, g_strdup ((char *)l->data));
+			}
 		}
 
 		g_list_foreach (data->image_paths, (GFunc)g_free, NULL);
@@ -826,6 +884,8 @@ estimate_size (const char *uri,
 
 	g_free (data);
 
+	profile_end ("end", NULL);
+
 	return size;
 }
 
@@ -847,6 +907,19 @@ setup_cancel_button (void)
 	/* FIXME: disconnect signal handlers? */
 }
 
+static void
+eject_disc_cb (GtkWidget *widget,
+	       gpointer   user_data)
+	       
+{
+	NautilusBurnerWriteData *data;
+
+	data = nautilus_burn_progress_get_write_data ();
+	if (data != NULL && data->drive != NULL) {
+		nautilus_burn_drive_eject ((NautilusBurnDrive *)data->drive);
+	}
+}
+
 static void
 write_another_disc_cb (GtkWidget *widget,
 		       gpointer   user_data)
@@ -857,7 +930,9 @@ write_another_disc_cb (GtkWidget *widget,
 	setup_cancel_button ();
 
 	data = nautilus_burn_progress_get_write_data ();
-	nautilus_burn_drive_eject ((NautilusBurnDrive *)data->drive);
+	if (data != NULL && data->drive != NULL) {
+		nautilus_burn_drive_eject ((NautilusBurnDrive *)data->drive);
+	}
 
 	res = nautilus_burn_progress_write_disc ();
 
@@ -884,7 +959,7 @@ setup_close_button (void)
 	g_signal_connect (close_button, "clicked", G_CALLBACK (gtk_main_quit), NULL);
 
 	data = nautilus_burn_progress_get_write_data ();
-	if (data) {
+	if (data != NULL) {
 		GtkWidget *write_another_button;
 		write_another_button = glade_xml_get_widget (xml, "write_another_button");
 		gtk_widget_show (write_another_button);
@@ -892,6 +967,15 @@ setup_close_button (void)
 		g_signal_handlers_disconnect_by_func (write_another_button, write_another_disc_cb, NULL);
 		g_signal_connect (write_another_button, "clicked", G_CALLBACK (write_another_disc_cb), NULL);
 	}
+
+	if (data != NULL) {
+		GtkWidget *button;
+		button = glade_xml_get_widget (xml, "eject_button");
+		gtk_widget_show (button);
+
+		g_signal_handlers_disconnect_by_func (button, eject_disc_cb, NULL);
+		g_signal_connect (button, "clicked", G_CALLBACK (eject_disc_cb), NULL);
+	}
 }
 
 static gboolean 
@@ -926,13 +1010,13 @@ overwrite_existing_file (GtkWindow  *filesel,
 }
 
 static char *
-select_iso_filename (void)
+select_iso_filename (const char *default_name)
 {
 	GtkWidget *chooser;
 	char      *filename;
 	int        response;
 
-	chooser = gtk_file_chooser_dialog_new_with_backend (_("Choose a filename for the cdrom image"),
+	chooser = gtk_file_chooser_dialog_new_with_backend (_("Choose a filename for the disc image"),
 							    NULL,
 							    GTK_FILE_CHOOSER_ACTION_SAVE,
 							    "unix",
@@ -940,7 +1024,7 @@ select_iso_filename (void)
 							    GTK_STOCK_OK, GTK_RESPONSE_OK,
 							    NULL);
 	gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (chooser),
-					   "image.iso");
+					   default_name);
  reselect:
 	filename = NULL;
 	response = gtk_dialog_run (GTK_DIALOG (chooser));
@@ -948,7 +1032,7 @@ select_iso_filename (void)
 	if (response == GTK_RESPONSE_OK) {
 		filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (chooser));
 		if (g_file_test (filename, G_FILE_TEST_EXISTS)) {
-			if (!overwrite_existing_file (GTK_WINDOW (chooser), filename)) {
+			if (! overwrite_existing_file (GTK_WINDOW (chooser), filename)) {
 				g_free (filename);
 				goto reselect;
 			}
@@ -960,50 +1044,6 @@ select_iso_filename (void)
 	return filename;
 }
 
-static gboolean
-media_type_matches (NautilusBurnMediaType type,
-		    gboolean              is_blank,
-		    gboolean              was_unmounted)
-{
-	if (was_unmounted) {
-		switch (type) {
-		case NAUTILUS_BURN_MEDIA_TYPE_CDRW:
-		case NAUTILUS_BURN_MEDIA_TYPE_DVDRW:
-		case NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_RW:
-			return TRUE;
-		default:
-			return FALSE;
-		}
-	}
-
-	switch (type) {
-	case NAUTILUS_BURN_MEDIA_TYPE_ERROR:
-	case NAUTILUS_BURN_MEDIA_TYPE_BUSY:
-		return FALSE;
-	case NAUTILUS_BURN_MEDIA_TYPE_UNKNOWN:
-		return TRUE;
-	case NAUTILUS_BURN_MEDIA_TYPE_CD:
-	case NAUTILUS_BURN_MEDIA_TYPE_DVD:
-	case NAUTILUS_BURN_MEDIA_TYPE_DVD_RAM:
-		return FALSE;
-	case NAUTILUS_BURN_MEDIA_TYPE_CDR:
-	case NAUTILUS_BURN_MEDIA_TYPE_DVDR:
-	case NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_R:
-		if (is_blank)
-			return TRUE;
-		else
-			return FALSE;
-	case NAUTILUS_BURN_MEDIA_TYPE_CDRW:
-	case NAUTILUS_BURN_MEDIA_TYPE_DVDRW:
-	case NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_RW:
-		return TRUE;
-	default:
-		break;
-	}
-
-	return FALSE;
-}
-
 static gboolean
 request_media (NautilusBurnDrive *drive,
 	       gint64             required_size,
@@ -1012,93 +1052,94 @@ request_media (NautilusBurnDrive *drive,
 {
 	gint64                media_size;
 	gboolean              reload;
-	gboolean              last_was_unmount;
 	NautilusBurnMediaType type;
 	gboolean              is_rewritable;
+	gboolean              can_write;
 	gboolean              is_blank;
 	gboolean              has_data;
 	gboolean              has_audio;
 
-	last_was_unmount = FALSE;
-	reload           = FALSE;
-
-	if (drive->type == NAUTILUS_BURN_DRIVE_TYPE_FILE)
+	/* If the output is a file no need to prompt */
+	if (drive->type == NAUTILUS_BURN_DRIVE_TYPE_FILE) {
 		return TRUE;
+	}
 
-	if (request_overburn)
+	reload = FALSE;
+
+ again:
+
+	/* if drive is mounted then unmount before checking anything */
+	if (nautilus_burn_drive_is_mounted (drive)) {
+		gboolean res;
+		res = nautilus_burn_drive_unmount (drive);
+		if (! res) {
+			g_warning ("Couldn't unmount volume in drive: %s", drive->device);
+		}
+	}
+
+	/* if overburn requested then no need to check size/type */
+	if (request_overburn) {
 		return TRUE;
+	}
 
 	media_size = nautilus_burn_drive_get_media_size (drive);
-
 	type = nautilus_burn_drive_get_media_type_full (drive, &is_rewritable, &is_blank, &has_data, &has_audio);
+	can_write = nautilus_burn_drive_media_type_is_writable (type, is_blank);
 
-	while ((!media_type_matches (type, is_blank, last_was_unmount))
-	       || (required_size > media_size)) {
-		gboolean              busy_cd;
-		gboolean              request_overburn;
-		int                   res;
-
-		busy_cd = (type == NAUTILUS_BURN_MEDIA_TYPE_BUSY);
-
-		if (busy_cd && !last_was_unmount) {
-			/* umount the disc if we find it, and try it again */
-			if (nautilus_burn_drive_unmount (drive)) {
-				last_was_unmount = TRUE;
-				goto retry;
-			}
-		}
+	if ((! can_write) || (required_size > media_size)) {
+		gboolean suggest_overburn;
+		int      res;
 
 		reload = (media_size > 0);
-
-		if (type == NAUTILUS_BURN_MEDIA_TYPE_ERROR)
+		if (type == NAUTILUS_BURN_MEDIA_TYPE_ERROR) {
 			reload = FALSE;
+		}
 
-		request_overburn = ((required_size - media_size) <= SUGGEST_OVERBURN);
+		suggest_overburn = ((required_size - media_size) <= SUGGEST_OVERBURN);
 
 		res = ask_for_media (drive,
 				     reload,
 				     FALSE,
-				     request_overburn,
+				     suggest_overburn,
 				     required_size);
 
-		if (res == GTK_RESPONSE_CANCEL)
+		if (res == GTK_RESPONSE_CANCEL) {
 			return FALSE;
+		}
 
 		if (res == RESPONSE_OVERBURN) {
-			if (use_overburn)
+			if (use_overburn) {
 				*use_overburn = TRUE;
-			break;
+			}
+
+			return TRUE;
 		}
 
-		last_was_unmount = FALSE;
-
-	retry:
-		type = nautilus_burn_drive_get_media_type_full (drive, &is_rewritable, &is_blank, &has_data, &has_audio);
-
-		reload = FALSE;
-		if (type == NAUTILUS_BURN_MEDIA_TYPE_UNKNOWN) {
-			reload = TRUE;
-		}
-
-		media_size = nautilus_burn_drive_get_media_size (drive);
-
-	}
-
-	if (!is_blank) {
-		/* don't ask user for confirmation of blank here, wait for warn-data-loss signal */
-		/*return ask_rewrite_disc (drive);*/
+		goto again;
 	}
 
 	return TRUE;
 }
 
 static char *
-get_output_filename (void)
+get_output_filename (const char *disc_name)
 {
-	char *filename = NULL;
+	char *filename         = NULL;
+	char *default_filename = NULL;
+
+	if (disc_name) {
+		default_filename = g_strdup_printf ("%s.iso", disc_name);
+	}
+
+	if (! default_filename) {
+		/* Translators: this is the filename of the image */
+		default_filename = g_strdup (_("image.iso"));
+	}
 
 	/* Run the file selector and get the iso file name selected */
-	filename = select_iso_filename ();
+	filename = select_iso_filename (default_filename);
+	g_free (default_filename);
+
 	if (! filename)
 		return NULL;
 
@@ -1208,18 +1249,152 @@ write_disc (NautilusBurnerWriteData *data)
 }
 
 static int
-burn_cd (const NautilusBurnDrive *drive,
-	 const NautilusBurnDrive *source_drive,
-	 const char              *source_iso,
-	 const char              *source_cue,
-	 const char              *label,
-	 gint64                   burn_size,
-	 gboolean                 eject_cd,
-	 int                      speed,
-	 gboolean                 dummy,
-	 gboolean                 debug,
-	 gboolean                 overburn,
-	 gboolean                 burnproof)
+get_input_type_from_source_data (NautilusBurnerSourceData *burn_data)
+{
+	int input_type;
+
+	if (! burn_data) {
+		input_type = NCB_SOURCE_BURN;
+	} else if (burn_data->source_iso || burn_data->source_cue) {
+		input_type = NCB_SOURCE_IMAGE;
+	} else if (burn_data->source_drive || burn_data->source_device) {
+		input_type = NCB_SOURCE_DRIVE;
+	} else {
+		input_type = NCB_SOURCE_BURN;
+	}
+
+	return input_type;
+}
+
+static void
+update_progress_dialog_disc_type (NautilusBurnDrive        *drive,
+				  NautilusBurnerSourceData *burn_data,
+				  NautilusBurnMediaType     type)
+{
+	GtkWidget *label;
+	GtkWidget *dialog;
+	GtkWidget *heading;
+	const char *label_text;
+	const char *title_text;
+	const char *heading_text;
+	char      *heading_markup;
+	int        input_type;
+	int        output_type;
+	/* Arrays structured as [input-type][output-type]
+	   input-type: files/default, image, drive
+	   output-type: disc/default, file, CD, DVD
+	 */
+	static const char *titles [][4] = {
+		{ N_("Writing Files to Disc"),
+		  N_("Writing Files to a Disc Image"),
+		  N_("Writing Files to CD"),
+		  N_("Writing Files to DVD")
+		},
+
+		{ N_("Writing Image to Disc"),
+		  N_("Copying Disc Image"),
+		  N_("Writing Image to CD"),
+		  N_("Writing Image to DVD")
+		},
+
+		{ N_("Copying Disc"),
+		  N_("Copying Disc to a Disc Image"),
+		  N_("Copying Disc to CD"),
+		  N_("Copying Disc to DVD")
+		}
+	};
+	static const char *headings [][4] = {
+		{ N_("Writing files to disc"),
+		  N_("Writing files to a disc image"),
+		  N_("Writing files to CD"),
+		  N_("Writing files to DVD")
+		},
+
+		{ N_("Writing image to disc"),
+		  N_("Copying disc image"),
+		  N_("Writing image to CD"),
+		  N_("Writing image to DVD")
+		},
+
+		{ N_("Copying disc"),
+		  N_("Copying disc to a disc image"),
+		  N_("Copying disc to CD"),
+		  N_("Copying disc to DVD")
+		}
+	};
+	static const char *descriptions [][4] = {
+		{ N_("The selected files are being written to a CD or DVD.  This operation may take a long time, depending on data size and write speed."),
+		  N_("The selected files are being written to a disc image file."),
+		  N_("The selected files are being written to a CD.  This operation may take a long time, depending on data size and write speed."),
+		  N_("The selected files are being written to a DVD.  This operation may take a long time, depending on data size and write speed.")
+		},
+
+		{ N_("The selected disc image is being written to a CD or DVD.  This operation may take a long time, depending on data size and write speed."),
+		  N_("The selected disc image is being copied to a disc image file."),
+		  N_("The selected disc image is being written to a CD.  This operation may take a long time, depending on data size and write speed."),
+		  N_("The selected disc image is being written to a DVD.  This operation may take a long time, depending on data size and write speed.")
+		},
+
+		{ N_("The selected disc is being copied to a CD or DVD.  This operation may take a long time, depending on data size and drive speed."),
+		  N_("The selected disc is being copied to a disc image file.  This operation may take a long time, depending on data size and drive speed."),
+		  N_("The selected disc is being copied to a CD.  This operation may take a long time, depending on data size and drive speed."),
+		  N_("The selected disc is being copied to a DVD.  This operation may take a long time, depending on data size and drive speed.")
+		}
+	};
+
+	input_type = get_input_type_from_source_data (burn_data);
+
+	if (drive->type == NAUTILUS_BURN_DRIVE_TYPE_FILE) {
+		output_type = NCB_OUTPUT_FILE;
+	} else {
+		switch (type) {
+		case NAUTILUS_BURN_MEDIA_TYPE_CD:
+		case NAUTILUS_BURN_MEDIA_TYPE_CDR:
+		case NAUTILUS_BURN_MEDIA_TYPE_CDRW:
+			output_type = NCB_OUTPUT_CD;
+			break;
+		case NAUTILUS_BURN_MEDIA_TYPE_DVD:
+		case NAUTILUS_BURN_MEDIA_TYPE_DVDR:
+		case NAUTILUS_BURN_MEDIA_TYPE_DVDRW:
+		case NAUTILUS_BURN_MEDIA_TYPE_DVD_RAM:
+		case NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_R:
+		case NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_RW:
+		case NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_R_DL:
+			output_type = NCB_OUTPUT_DVD;
+			break;
+		case NAUTILUS_BURN_MEDIA_TYPE_UNKNOWN:
+		default:
+			output_type = NCB_OUTPUT_DEFAULT;
+			break;
+		}
+	}
+
+	label_text = descriptions [input_type][output_type];
+	title_text = titles [input_type][output_type];
+	heading_text = headings [input_type][output_type];
+
+	label   = glade_xml_get_widget (xml, "progress_description_label");
+	heading = glade_xml_get_widget (xml, "info_title_label");
+	dialog  = glade_xml_get_widget (xml, "progress_window");
+
+	gtk_label_set_text (GTK_LABEL (label), _(label_text));
+	gtk_window_set_title (GTK_WINDOW (dialog), _(title_text));
+
+	heading_markup = g_strdup_printf ("<big><b>%s</b></big>", _(heading_text));
+	gtk_label_set_markup (GTK_LABEL (heading), heading_markup);
+	g_free (heading_markup);
+}
+
+static int
+burn_cd (NautilusBurnDrive        *drive,
+	 NautilusBurnerSourceData *burn_data,
+	 const char               *label,
+	 gboolean                  eject_cd,
+	 int                       speed,
+	 gboolean                  dummy,
+	 gboolean                  debug,
+	 gboolean                  overburn,
+	 gboolean                  burnproof)
 {
 	GError                        *error        = NULL;
 	char                          *filename     = NULL;
@@ -1229,11 +1404,28 @@ burn_cd (const NautilusBurnDrive *drive,
 	int                            fd;
 	int                            res      = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
 	gboolean                       try_overburn = FALSE;
+	NautilusBurnDrive             *source_drive;
+	const char                    *source_iso;
+	const char                    *source_cue;
+	gint64                         burn_size;
+	NautilusBurnMediaType          media_type = NAUTILUS_BURN_MEDIA_TYPE_UNKNOWN;
+
+	source_drive = burn_data->source_drive;
+	source_iso   = burn_data->source_iso;
+	source_cue   = burn_data->source_cue;
+	burn_size    = burn_data->size;
 
 	filename = NULL;
 
 	if (drive->type == NAUTILUS_BURN_DRIVE_TYPE_FILE) {
-		filename = get_output_filename ();
+		char *disc_name = NULL;
+
+		if (source_drive)
+			disc_name = nautilus_burn_drive_get_media_label (source_drive);
+
+		filename = get_output_filename (disc_name);
+		g_free (disc_name);
+
 		if (! filename)
 			return NAUTILUS_BURN_RECORDER_RESULT_CANCEL;
 	} else {
@@ -1265,23 +1457,29 @@ burn_cd (const NautilusBurnDrive *drive,
 	if (drive->type != NAUTILUS_BURN_DRIVE_TYPE_FILE) {
 		/* if we are reading from a drive then don't request media yet */
 		if (! source_drive) {
-			gboolean is_locked;
-			gboolean got_media;
+			gboolean              is_locked;
+			gboolean              got_media;
+
+			is_locked = nautilus_burn_drive_lock (drive, _("Burning CD"), NULL);
+			got_media = request_media (drive, burn_size, overburn, &try_overburn);
+
+			if (got_media)
+				media_type = nautilus_burn_drive_get_media_type (drive);
 
-			is_locked = nautilus_burn_drive_lock ((NautilusBurnDrive *)drive, _("Burning CD"), NULL);
-			got_media = request_media ((NautilusBurnDrive *)drive, burn_size, overburn, &try_overburn);
 			if (is_locked) {
-				nautilus_burn_drive_unlock ((NautilusBurnDrive *)drive);
+				nautilus_burn_drive_unlock (drive);
 			} 
 
-			if (!got_media) {
+			if (! got_media) {
 				res = NAUTILUS_BURN_RECORDER_RESULT_CANCEL;
 				goto out;
 			}
-
 		}
 	}
 
+	/* update the description now that we may know the media type */
+	update_progress_dialog_disc_type (drive, burn_data, media_type);
+
 	if (source_iso == NULL
 	    && source_cue == NULL) {
 		NautilusBurnImageCreateFlags image_flags = 0;
@@ -1299,15 +1497,16 @@ burn_cd (const NautilusBurnDrive *drive,
 				image_flags |= NAUTILUS_BURN_IMAGE_CREATE_WARN_SPACE;
 
 			/* Don't warn on low space unless the user selected this file specifically */
-			if (source_drive)
+			if (source_drive) {
 				res = nautilus_burn_make_iso_from_drive (filename,
 									 source_drive,
 									 image_flags,
 									 &image_type,
 									 &toc_filename,
 									 &error);
-			else
+			} else {
 				res = nautilus_burn_make_iso (filename, label, image_flags, &image_type, &error);
+			}
 		} else {
 			res = NAUTILUS_BURN_RECORDER_RESULT_RETRY;
 		}
@@ -1355,15 +1554,16 @@ burn_cd (const NautilusBurnDrive *drive,
 
 				image_flags |= NAUTILUS_BURN_IMAGE_CREATE_WARN_SPACE;
 
-				if (source_drive)
+				if (source_drive) {
 					res = nautilus_burn_make_iso_from_drive (filename,
 										 source_drive,
 										 image_flags,
 										 &image_type,
 										 &toc_filename,
 										 &error);
-				else
+				} else {
 					res = nautilus_burn_make_iso (filename, label, image_flags, &image_type, &error);
+				}
 			}
 		}
 
@@ -1434,13 +1634,16 @@ burn_cd (const NautilusBurnDrive *drive,
 		if (1) {
 			record_flags |= NAUTILUS_BURN_RECORDER_WRITE_BURNPROOF;
 		}
-
-		if (source_drive
-		    && strcmp (source_drive->device, drive->device) == 0) {
-			nautilus_burn_drive_eject ((NautilusBurnDrive *)source_drive);
+		if (1) {
+			record_flags |= NAUTILUS_BURN_RECORDER_WRITE_DISC_AT_ONCE;
 		}
 
-		data->drive  = nautilus_burn_drive_copy ((NautilusBurnDrive *)drive);
+		if (source_drive
+		    && nautilus_burn_drive_equal (source_drive, drive)) {
+			nautilus_burn_drive_eject (source_drive);
+		}
+
+		data->drive  = nautilus_burn_drive_ref (drive);
 		data->tracks = g_list_prepend (data->tracks, track);
 		data->flags  = record_flags;
 		data->speed  = speed;
@@ -1460,18 +1663,19 @@ burn_cd (const NautilusBurnDrive *drive,
 	return res;
 }
 
-static const NautilusBurnDrive *
+static NautilusBurnDrive *
 lookup_current_recorder (GladeXML *xml)
 {
-	GtkWidget               *selection;
-	const NautilusBurnDrive *drive;
+	GtkWidget         *selection;
+	NautilusBurnDrive *drive;
 
 	selection = glade_xml_get_widget (xml, "target_optionmenu");
 
 	if (! selection)
 		return NULL;
 
-	drive = nautilus_burn_drive_selection_get_drive (NAUTILUS_BURN_DRIVE_SELECTION (selection));
+	/* the drive is reffed */
+	drive = nautilus_burn_drive_selection_get_active (NAUTILUS_BURN_DRIVE_SELECTION (selection));
 
 	return drive;
 }
@@ -1498,13 +1702,16 @@ static void
 refresh_dialog (GladeXML *xml)
 {
 	GtkWidget               *combobox;
-	GtkWidget               *eject_check;
 	GConfClient             *gc;
 	char                    *name;
-	int                      i, default_speed;
-	const NautilusBurnDrive *drive;
+	int                      i;
+	int                      default_speed;
+	int                      default_speed_index;
+	NautilusBurnDrive       *drive;
 	GtkTreeModel            *model;
-        GtkTreeIter              iter;
+	GtkTreeIter              iter;
+
+	default_speed_index = 0;
 
 	/* Find active recorder: */
 	drive = lookup_current_recorder (xml);
@@ -1525,26 +1732,34 @@ refresh_dialog (GladeXML *xml)
 			    -1);
 
 	if (drive) {
-		for (i = 1; i <= drive->max_speed_write; i++) {
-			name = g_strdup_printf ("%d \303\227", i);
+		const int *write_speeds;
 
+		write_speeds = nautilus_burn_drive_get_write_speeds (drive);
+
+		for (i = 0; write_speeds[i] > 0; i++) {
+
+			if (write_speeds[i] == default_speed) {
+				default_speed_index = i + 1;
+			}
+
+			name = g_strdup_printf ("%d \303\227", write_speeds[i]);
 			gtk_list_store_append (GTK_LIST_STORE (model), &iter);
 			gtk_list_store_set (GTK_LIST_STORE (model), &iter,
 					    0, name,
-					    1, i,
+					    1, write_speeds [i],
 					    -1);
 			g_free (name);
 		}
 
-		/* Disable speed if max speed < 1 */
-		gtk_widget_set_sensitive (combobox, drive->max_speed_write > 0);
-		/* Disable eject for burn-to-iso */
-		eject_check = glade_xml_get_widget (xml, "eject_check");
-		gtk_widget_set_sensitive (eject_check, drive->type != NAUTILUS_BURN_DRIVE_TYPE_FILE);
+		/* Disable speed if no items in list */
+		gtk_widget_set_sensitive (combobox, i > 0);
 	}
 
-	/* for now assume equivalence between index in comboxbox and speed */
-	gtk_combo_box_set_active (GTK_COMBO_BOX (combobox), default_speed);
+	gtk_combo_box_set_active (GTK_COMBO_BOX (combobox), default_speed_index);
+
+	if (drive) {
+		nautilus_burn_drive_unref (drive);
+	}
 }
 
 static void
@@ -1704,9 +1919,9 @@ ask_cancel (void)
 
 	image = gtk_image_new_from_stock (GTK_STOCK_STOP, GTK_ICON_SIZE_BUTTON);
 	gtk_widget_show (image);
-	gtk_dialog_add_button (GTK_DIALOG (warning_dialog), _("Continue"), GTK_RESPONSE_NO);
 	button = gtk_dialog_add_button (GTK_DIALOG (warning_dialog), _("Interrupt"), GTK_RESPONSE_YES);
 	g_object_set (button, "image", image, NULL);
+	gtk_dialog_add_button (GTK_DIALOG (warning_dialog), _("Continue"), GTK_RESPONSE_NO);
 
 	gtk_dialog_set_default_response (GTK_DIALOG (warning_dialog), GTK_RESPONSE_NO);
 	res = gtk_dialog_run (GTK_DIALOG (warning_dialog));
@@ -1956,19 +2171,16 @@ selection_dialog_response (GtkWidget                *dialog,
 	}
 
 	if (response_id == GTK_RESPONSE_OK) {
-		const NautilusBurnDrive *drive;
+		NautilusBurnDrive *drive;
 		GtkWidget  *progress_window;
 		GtkWidget  *cancel_button;
 		GtkWidget  *entry;
 		GtkWidget  *speed_combobox;
-		GtkWidget  *eject_check;
 		GtkWidget  *dummy_check;
 		GtkWidget  *description_label;
 		gboolean    eject;
 		gboolean                 dummy, debug, overburn, burnproof;
 		const char              *label;
-		const char              *label_format;
-		char                    *label_text;
 		int                      res, speed;
 		GConfClient             *gconf_client;
 
@@ -1979,17 +2191,12 @@ selection_dialog_response (GtkWidget                *dialog,
 		cancel_button      = glade_xml_get_widget (xml, "cancel_button");
 		description_label  = glade_xml_get_widget (xml, "progress_description_label");
 		entry              = glade_xml_get_widget (xml, "cdname_entry");
-		eject_check        = glade_xml_get_widget (xml, "eject_check");
 		speed_combobox     = glade_xml_get_widget (xml, "speed_combobox");
 		dummy_check        = glade_xml_get_widget (xml, "dummy_check");
+		progress_window    = glade_xml_get_widget (xml, "progress_window");
 
-		label_format = gtk_label_get_text (GTK_LABEL (description_label));
-		/* FIXME: get type of media in drive */
-		label_text = g_strdup_printf (label_format, _("CD or DVD"));
-		gtk_label_set_text (GTK_LABEL (description_label), label_text);
-		g_free (label_text);
+		update_progress_dialog_disc_type (drive, burn_data, NAUTILUS_BURN_MEDIA_TYPE_UNKNOWN);
 
-		progress_window = glade_xml_get_widget (xml, "progress_window");
 		gtk_window_set_icon_name (GTK_WINDOW (progress_window), "gnome-dev-cdrom");
 		nautilus_burn_progress_image_setup ();
 		gtk_widget_show (progress_window);
@@ -1998,7 +2205,10 @@ selection_dialog_response (GtkWidget                *dialog,
 		g_signal_connect (progress_window, "delete_event", G_CALLBACK (handle_delete_event), NULL);
 
 		label = gtk_entry_get_text (GTK_ENTRY (entry));
-		eject = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (eject_check));
+
+		/* eject all discs */
+		/* FIXME: make this gconf configurable? */
+		eject = FALSE;
 
 		speed = get_speed_selection (speed_combobox);
 
@@ -2015,11 +2225,8 @@ selection_dialog_response (GtkWidget                *dialog,
 		}
 
 		res = burn_cd (drive,
-			       burn_data->source_drive,
-			       burn_data->source_iso,
-			       burn_data->source_cue,
+			       burn_data,
 			       label,
-			       burn_data->size,
 			       eject,
 			       speed,
 			       dummy,
@@ -2027,6 +2234,9 @@ selection_dialog_response (GtkWidget                *dialog,
 			       overburn,
 			       burnproof);
 
+		if (drive)
+			nautilus_burn_drive_unref (drive);
+
 		if (res == NAUTILUS_BURN_RECORDER_RESULT_CANCEL) {
 			gtk_main_quit ();
 		}
@@ -2056,8 +2266,8 @@ fill_selection_dialog (NautilusBurnerSourceData *burn_data)
 
 			goto done;
 		}
-
 		nautilus_burn_drive_unmount (burn_data->source_drive);
+		disc_name = nautilus_burn_drive_get_media_label (burn_data->source_drive);
 		burn_data->size = nautilus_burn_drive_get_media_size (burn_data->source_drive);
 
 		if (burn_data->size <= 0) {
@@ -2104,11 +2314,30 @@ fill_selection_dialog (NautilusBurnerSourceData *burn_data)
 	return FALSE;
 
  done:
+	g_free (disc_name);
+
 	gtk_main_quit ();
 
 	return FALSE;
 }
 
+static void
+update_selection_dialog_source_type (NautilusBurnerSourceData *burn_data)
+{
+	GtkWidget *widget;
+	int        input_type;
+
+	input_type = get_input_type_from_source_data (burn_data);
+
+	if (NCB_SOURCE_DRIVE == input_type) {
+		widget = glade_xml_get_widget (xml, "cdr_dialog");
+		gtk_window_set_title (GTK_WINDOW (widget), _("Copy Disc"));
+
+		widget = glade_xml_get_widget (xml, "target_label");
+		gtk_label_set_text_with_mnemonic (GTK_LABEL (widget), _("Copy disc _to:"));
+	}
+}
+
 static gboolean
 create_selection_dialog (NautilusBurnerSourceData *burn_data)
 {
@@ -2144,6 +2373,8 @@ create_selection_dialog (NautilusBurnerSourceData *burn_data)
 	/* make insensitive until fill_dialog finishes */
 	gtk_widget_set_sensitive (dialog, FALSE);
 
+	update_selection_dialog_source_type (burn_data);
+
 	gtk_widget_show (dialog);
 
 	cursor = gdk_cursor_new (GDK_WATCH);
@@ -2157,6 +2388,28 @@ create_selection_dialog (NautilusBurnerSourceData *burn_data)
 	return FALSE;
 }
 
+static char *
+expand_path_input (const char *input)
+{
+	char *uri;
+	char *path;
+
+	uri = gnome_vfs_make_uri_from_input (input);
+	if (uri == NULL) {
+		g_warning ("Could not parse filename");
+		return NULL;
+	}
+
+	path = gnome_vfs_get_local_path_from_uri (uri);
+	g_free (uri);
+
+	if (path == NULL) {
+		g_warning ("Only local files are supported at this time");
+	}
+
+	return path;
+}
+
 int
 main (int argc, char *argv[])
 {
@@ -2203,14 +2456,34 @@ main (int argc, char *argv[])
 	}
 
 	burn_data = g_new0 (NautilusBurnerSourceData, 1);
-	burn_data->source_cue = g_strdup (source_cue);
-	burn_data->source_iso = g_strdup (source_iso);
+
+	if (source_iso != NULL) {
+		char *path;
+
+		path = expand_path_input (source_iso);
+		if (path == NULL) {
+			goto out;
+		}
+		burn_data->source_iso = path;
+	}
+
+	if (source_cue != NULL) {
+		char *path;
+
+		path = expand_path_input (source_cue);
+		if (path == NULL) {
+			goto out;
+		}
+		burn_data->source_cue = path;
+	}
+
 	burn_data->source_device = g_strdup (source_device);
 
 	g_idle_add ((GSourceFunc)create_selection_dialog, burn_data);
 
 	gtk_main ();
 
+ out:
 	g_free (burn_data->source_cue);
 	g_free (burn_data->source_iso);
 	g_free (burn_data->source_device);
